from collections.abc import Iterable
from typing import override

import jax.numpy as jnp
from jaxtyping import Array, Float

from liblaf.peach import tree
from liblaf.peach.optim.objective import Objective

from ._abc import LineSearch

type Scalar = Float[Array, ""]
type Vector = Float[Array, " N"]


@tree.define
class LineSearchMin(LineSearch):
    methods: Iterable[LineSearch] = tree.field()

    @override
    def search(
        self,
        objective: Objective,
        params: Vector,
        grad: Vector,
        search_direction: Vector,
    ) -> Scalar:
        return jnp.min(
            jnp.stack(
                [
                    method.search(objective, params, grad, search_direction)
                    for method in self.methods
                ]
            )
        )



import abc

from jaxtyping import Array, Float

from liblaf.peach import tree
from liblaf.peach.optim.objective import Objective

type Scalar = Float[Array, ""]
type Vector = Float[Array, " N"]


@tree.define
class LineSearch:
    @abc.abstractmethod
    def search(
        self,
        objective: Objective,
        params: Vector,
        grad: Vector,
        search_direction: Vector,
    ) -> Scalar:
        raise NotImplementedError



from jaxtyping import Array, Float

from liblaf.peach import tree
from liblaf.peach.linalg.abc import State, Stats

type Scalar = Float[Array, ""]


@tree.define
class CompositeState(State):
    state: list[State] = tree.field(factory=list)


@tree.define
class CompositeStats(Stats):
    stats: list[Stats] = tree.field(factory=list)
    relative_residual: Scalar = tree.array(default=None, kw_only=True)



from typing import override

import jax.numpy as jnp
from jaxtyping import Array, Float

from liblaf.peach import tree
from liblaf.peach.optim.objective import Objective

from ._abc import LineSearch

type Scalar = Float[Array, ""]
type Vector = Float[Array, " N"]


@tree.define
class LineSearchCollisionRepulsionThreshold(LineSearch):
    collision_repulsion_threshold: Scalar = tree.array(
        default=jnp.inf, converter=tree.converters.asarray
    )

    @override
    def search(
        self,
        objective: Objective,
        params: Vector,
        grad: Vector,
        search_direction: Vector,
    ) -> Scalar:
        p_norm: Scalar = jnp.linalg.norm(search_direction, ord=jnp.inf)
        return self.collision_repulsion_threshold / (2.0 * p_norm)  # pyright: ignore[reportAssignmentType]



from jaxtyping import Array, Float

from liblaf.peach import tree
from liblaf.peach.optim.abc import Stats

type Scalar = Float[Array, ""]


@tree.define
class PNCGStats(Stats):
    relative_decrease: Scalar = tree.array(default=None, kw_only=True)



from . import abc, linesearch, objective, optax, pncg, scipy
from .abc import Callback, Optimizer, OptimizeSolution, Result
from .linesearch import (
    LineSearch,
    LineSearchCollisionRepulsionThreshold,
    LineSearchMin,
    LineSearchNaive,
    LineSearchSingleNewton,
)
from .objective import Objective
from .optax import Optax
from .pncg import PNCG
from .scipy import ScipyOptimizer

__all__ = [
    "PNCG",
    "Callback",
    "LineSearch",
    "LineSearchCollisionRepulsionThreshold",
    "LineSearchMin",
    "LineSearchNaive",
    "LineSearchSingleNewton",
    "Objective",
    "Optax",
    "OptimizeSolution",
    "Optimizer",
    "Result",
    "ScipyOptimizer",
    "abc",
    "linesearch",
    "objective",
    "optax",
    "pncg",
    "scipy",
]



from __future__ import annotations

import functools
from collections.abc import Callable
from typing import Any, TypedDict, Unpack, dataclass_transform, overload

import attrs

from liblaf import grapes

from ._field import array, container, field
from ._register_fieldz import register_fieldz


class DefineKwargs(TypedDict, total=False):
    these: dict[str, Any] | None
    repr: bool
    unsafe_hash: bool | None
    hash: bool | None
    init: bool
    slots: bool
    frozen: bool
    weakref_slot: bool
    str: bool
    auto_attribs: bool
    kw_only: bool
    cache_hash: bool
    auto_exc: bool
    eq: bool | None
    order: bool | None
    auto_detect: bool
    getstate_setstate: bool | None
    on_setattr: attrs._OnSetAttrArgType | None
    field_transformer: attrs._FieldTransformer | None
    match_args: bool


@dataclass_transform(field_specifiers=(attrs.field, array, container, field))
@overload
def define[T: type](cls: T, /, **kwargs: Unpack[DefineKwargs]) -> T: ...
@overload
def define[T: type](
    cls: None = None, /, **kwargs: Unpack[DefineKwargs]
) -> Callable[[T], T]: ...
def define(cls: type | None = None, /, **kwargs: Unpack[DefineKwargs]) -> Any:
    if cls is None:
        return functools.partial(define, **kwargs)
    cls = grapes.attrs.define(cls, **kwargs)
    cls = register_fieldz(cls)
    return cls



import logging
from typing import override

import jax.numpy as jnp
from jaxtyping import Array, Float, Integer

from liblaf.peach import tree
from liblaf.peach.optim.objective import Objective

from ._abc import LineSearch

type Scalar = Float[Array, ""]
type Vector = Float[Array, " N"]


logger: logging.Logger = logging.getLogger(__name__)


@tree.define
class LineSearchNaive(LineSearch):
    initial: LineSearch = tree.field()

    decay: Scalar = tree.array(
        default=0.5, converter=tree.converters.asarray, kw_only=True
    )
    max_steps: Integer[Array, ""] = tree.array(
        default=20, converter=tree.converters.asarray, kw_only=True
    )

    @override
    def search(
        self,
        objective: Objective,
        params: Vector,
        grad: Vector,
        search_direction: Vector,
    ) -> Scalar:
        step_size: Scalar = self.initial.search(
            objective, params, grad, search_direction
        )
        if self.max_steps == 0:
            return step_size
        assert objective.fun is not None
        f0: Scalar = objective.fun(params)
        for _ in range(self.max_steps):
            params_next: Vector = params + step_size * search_direction
            f_next: Scalar = objective.fun(params_next)
            if jnp.isfinite(f_next) and f_next < f0:
                break
            step_size *= self.decay
        return step_size



from typing import overload

import jax.numpy as jnp
from jax import Array
from jaxtyping import ArrayLike


@overload
def asarray(value: None) -> None: ...
@overload
def asarray(value: ArrayLike) -> Array: ...
def asarray(value: ArrayLike | None) -> Array | None:
    if value is None:
        return None
    return jnp.asarray(value)



from typing import override

import jax.numpy as jnp
from jaxtyping import Array, Float

from liblaf.peach import tree
from liblaf.peach.optim.objective import Objective

from ._abc import LineSearch

type Scalar = Float[Array, ""]
type Vector = Float[Array, " N"]


@tree.define
class LineSearchSingleNewton(LineSearch):
    @override
    def search(
        self,
        objective: Objective,
        params: Vector,
        grad: Vector,
        search_direction: Vector,
    ) -> Scalar:
        assert objective.hess_quad is not None
        hess_quad: Scalar = objective.hess_quad(params, search_direction)
        if not jnp.isfinite(hess_quad) or hess_quad <= 0.0:
            hess_quad = jnp.sum(jnp.square(search_direction))
        return -jnp.vdot(grad, search_direction) / hess_quad



# ruff: noqa: N803, N806

from __future__ import annotations

from collections.abc import Iterable
from typing import override

import equinox as eqx
import jax.numpy as jnp
from jaxtyping import Array, Bool, Float, Integer

from liblaf.peach import tree
from liblaf.peach.constraints import Constraint
from liblaf.peach.optim.abc import (
    Optimizer,
    OptimizeSolution,
    Params,
    Result,
    SetupResult,
)
from liblaf.peach.optim.objective import Objective

from ._state import PNCGState
from ._stats import PNCGStats

type Scalar = Float[Array, ""]
type Vector = Float[Array, " N"]


@tree.define
class PNCG(Optimizer[PNCGState, PNCGStats]):
    from ._state import PNCGState as State
    from ._stats import PNCGStats as Stats

    Solution = OptimizeSolution[State, Stats]

    max_steps: Integer[Array, ""] = tree.array(
        default=256, converter=tree.converters.asarray, kw_only=True
    )
    atol: Scalar = tree.array(
        default=1e-15, converter=tree.converters.asarray, kw_only=True
    )
    rtol: Scalar = tree.array(
        default=1e-5, converter=tree.converters.asarray, kw_only=True
    )

    stagnation_patience: Integer[Array, ""] = tree.array(
        default=20, converter=tree.converters.asarray, kw_only=True
    )
    stagnation_max_restarts: Integer[Array, ""] = tree.array(
        default=5, converter=tree.converters.asarray, kw_only=True
    )

    beta_non_negative: Bool[Array, ""] = tree.array(
        default=False, converter=tree.converters.asarray, kw_only=True
    )
    beta_restart_threshold: Scalar = tree.array(
        default=jnp.inf, converter=tree.converters.asarray, kw_only=True
    )
    max_delta: Scalar = tree.array(
        default=jnp.inf, converter=tree.converters.asarray, kw_only=True
    )

    @override
    def init(
        self,
        objective: Objective,
        params: Params,
        *,
        constraints: Iterable[Constraint] = (),
    ) -> SetupResult[State, Stats]:
        params_flat: Vector
        objective, params_flat, constraints = objective.flatten(
            params, constraints=constraints
        )
        if self.jit:
            objective = objective.jit()
        if self.timer:
            objective = objective.timer()
        assert objective.structure is not None
        state: PNCG.State = self.State(
            params_flat=params_flat,
            structure=objective.structure,
            best_params_flat=params_flat,
        )
        return SetupResult(objective, constraints, state, self.Stats())

    @override
    def postprocess(
        self,
        objective: Objective,
        state: State,
        stats: Stats,
        result: Result,
        *,
        constraints: Iterable[Constraint] = (),
    ) -> OptimizeSolution[State, Stats]:
        state.params_flat = state.best_params_flat
        return super().postprocess(
            objective, state, stats, result, constraints=constraints
        )

    @override
    def step(
        self,
        objective: Objective,
        state: State,
        *,
        constraints: Iterable[Constraint] = (),
    ) -> State:
        if constraints:
            raise NotImplementedError
        assert objective.grad_and_hess_diag is not None
        assert objective.hess_quad is not None
        g: Vector
        H_diag: Vector
        g, H_diag = objective.grad_and_hess_diag(state.params_flat)
        H_diag = jnp.where(H_diag <= 0.0, 1.0, H_diag)
        P: Vector = jnp.reciprocal(H_diag)
        beta: Scalar
        p: Vector
        if state.search_direction_flat is None:
            beta = jnp.zeros(())
            p = -P * g
        elif state.stagnation_counter >= self.stagnation_patience:
            state.stagnation_counter = jnp.zeros_like(state.stagnation_counter)
            state.stagnation_restarts += 1
            beta = jnp.zeros(())
            p = -P * g
        else:
            beta = self._compute_beta(
                g_prev=state.grad_flat, g=g, p=state.search_direction_flat, P=P
            )
            p = -P * g + beta * state.search_direction_flat
        pHp: Scalar = objective.hess_quad(state.params_flat, p)
        alpha: Scalar = self._compute_alpha(g, p, pHp)
        delta_x: Vector = alpha * p
        delta_x = jnp.clip(delta_x, -self.max_delta, self.max_delta)
        state.params_flat += delta_x
        DeltaE: Scalar = -alpha * jnp.vdot(g, p) - 0.5 * alpha**2 * pHp
        if state.first_decrease is None:
            state.first_decrease = DeltaE
        grad_norm: Scalar = jnp.linalg.norm(g)
        if grad_norm > state.best_grad_norm:
            state.stagnation_counter += 1
        else:
            state.best_grad_norm = grad_norm
            state.best_params_flat = state.params_flat
            state.stagnation_counter = jnp.zeros_like(state.stagnation_counter)
        state.alpha = alpha
        state.beta = beta
        state.decrease = DeltaE
        state.grad_flat = g
        state.hess_diag_flat = H_diag
        state.hess_quad = pHp
        state.preconditioner_flat = P
        state.search_direction_flat = p
        return state

    @override
    def terminate(
        self,
        objective: Objective,
        state: PNCGState,
        stats: PNCGStats,
        *,
        constraints: Iterable[Constraint] = (),
    ) -> tuple[bool, Result]:
        if constraints:
            raise NotImplementedError
        assert state.first_decrease is not None
        stats.relative_decrease = state.decrease / state.first_decrease
        if (
            not jnp.isfinite(state.decrease)
            or (state.alpha is not None and not jnp.isfinite(state.alpha))
            or (state.beta is not None and not jnp.isfinite(state.beta))
        ):
            return False, Result.NAN
        if state.decrease < self.atol + self.rtol * state.first_decrease:
            return True, Result.SUCCESS
        if stats.n_steps >= self.max_steps:
            return True, Result.MAX_STEPS_REACHED
        if state.stagnation_restarts >= self.stagnation_max_restarts:
            return True, Result.STAGNATION
        return False, Result.UNKNOWN_ERROR

    @eqx.filter_jit
    def _compute_alpha(self, g: Vector, p: Vector, pHp: Scalar) -> Scalar:
        alpha: Scalar = -jnp.vdot(g, p) / pHp
        alpha = jnp.nan_to_num(alpha, nan=1.0)
        return alpha

    @eqx.filter_jit
    def _compute_beta(self, g_prev: Vector, g: Vector, p: Vector, P: Vector) -> Scalar:
        y: Vector = g - g_prev
        yTp: Scalar = jnp.vdot(y, p)
        Py: Scalar = P * y
        beta: Scalar = jnp.vdot(g, Py) / yTp - (jnp.vdot(y, Py) / yTp) * (
            jnp.vdot(p, g) / yTp
        )
        beta = jnp.nan_to_num(beta, nan=0.0)
        beta = jnp.where(self.beta_non_negative, jnp.maximum(beta, 0.0), beta)
        beta = jnp.where(self.beta_restart_threshold < beta, 0.0, beta)
        return beta



import warnings
from collections.abc import Iterable
from typing import override

import jax.numpy as jnp
import optax
from jaxtyping import Array, Float, Integer

from liblaf.peach import tree
from liblaf.peach.constraints import Constraint
from liblaf.peach.optim.abc import (
    Callback,
    Optimizer,
    OptimizeSolution,
    Params,
    Result,
    SetupResult,
)
from liblaf.peach.optim.objective import Objective

from ._types import OptaxState, OptaxStats

type Scalar = Float[Array, ""]
type Vector = Float[Array, " N"]


@tree.define
class Optax(Optimizer[OptaxState, OptaxStats]):
    from ._types import OptaxState as State
    from ._types import OptaxStats as Stats

    Solution = OptimizeSolution[State, Stats]
    Callback = Callback[State, Stats]

    wrapped: optax.GradientTransformation

    patience: Integer[Array, ""] = tree.array(
        default=20, converter=tree.converters.asarray, kw_only=True
    )
    rtol: Scalar = tree.array(
        default=0.0, converter=tree.converters.asarray, kw_only=True
    )

    @override
    def init(
        self,
        objective: Objective,
        params: Params,
        *,
        constraints: Iterable[Constraint] = (),
    ) -> SetupResult[State, Stats]:
        state: OptaxState
        stats: OptaxStats
        objective, constraints, state, stats = super().init(
            objective, params, constraints=constraints
        )
        state.wrapped = self.wrapped.init(state.params_flat)
        return SetupResult(objective, constraints, state, stats)

    @override
    def step(
        self,
        objective: Objective,
        state: State,
        *,
        constraints: Iterable[Constraint] = (),
    ) -> State:
        if constraints:
            warnings.warn(
                "Optax optimizer does not support constraints. Ignoring them.",
                RuntimeWarning,
                stacklevel=3,
            )
        assert objective.value_and_grad is not None
        state.value, state.grad_flat = objective.value_and_grad(state.params_flat)
        state.updates_flat, state.wrapped = self.wrapped.update(  # pyright: ignore[reportAttributeAccessIssue]
            state.grad_flat, state.wrapped, state.params_flat
        )
        state.params_flat = optax.apply_updates(state.params_flat, state.updates_flat)  # pyright: ignore[reportAttributeAccessIssue]
        return state

    @override
    def terminate(
        self,
        objective: Objective,
        state: State,
        stats: Stats,
        *,
        constraints: Iterable[Constraint] = (),
    ) -> tuple[bool, Result]:
        if state.value <= state.best_value_so_far:
            state.best_params_flat = state.params_flat
            state.best_value_so_far = state.value
            state.steps_from_best = jnp.zeros_like(state.steps_from_best)
        else:
            state.steps_from_best += 1
        if (
            state.steps_from_best > self.patience
            and state.value >= state.best_value_so_far * (1.0 - self.rtol)
        ):
            return True, Result.SUCCESS
        return False, Result.UNKNOWN_ERROR

    @override
    def postprocess(
        self,
        objective: Objective,
        state: OptaxState,
        stats: OptaxStats,
        result: Result,
        *,
        constraints: Iterable[Constraint] = (),
    ) -> OptimizeSolution[OptaxState, OptaxStats]:
        state.params_flat = state.best_params_flat
        return super().postprocess(
            objective, state, stats, result, constraints=constraints
        )



from ._abc import LineSearch
from ._collision_repulsion_threshold import LineSearchCollisionRepulsionThreshold
from ._min import LineSearchMin
from ._naive import LineSearchNaive
from ._single_newton import LineSearchSingleNewton

__all__ = [
    "LineSearch",
    "LineSearchCollisionRepulsionThreshold",
    "LineSearchMin",
    "LineSearchNaive",
    "LineSearchSingleNewton",
]



import jax.numpy as jnp
import optax
from jaxtyping import Array, Float, Integer

from liblaf.peach import tree
from liblaf.peach.optim.abc import Params, State, Stats
from liblaf.peach.tree import TreeView

type Scalar = Float[Array, ""]
type Vector = Float[Array, " N"]


@tree.define
class OptaxState(State):
    wrapped: optax.OptState = tree.field(default=None, kw_only=True)

    value: Scalar = tree.array(default=None, kw_only=True)

    grad = TreeView[Params]()
    grad_flat: Vector = tree.array(default=None, kw_only=True)

    updates = TreeView[Params]()
    updates_flat: Vector = tree.array(default=None, kw_only=True)

    best_params = TreeView[Params]()
    best_params_flat: Vector = tree.array(default=None, kw_only=True)
    best_value_so_far: Scalar = tree.array(default=jnp.inf, kw_only=True)
    steps_from_best: Integer[Array, ""] = tree.array(default=0, kw_only=True)


@tree.define
class OptaxStats(Stats):
    pass



from __future__ import annotations

from collections.abc import Callable, Mapping
from typing import Any, TypedDict, Unpack, overload

import attrs
import toolz
from jaxtyping import Array, ArrayLike
from liblaf.grapes import wraps

from liblaf.peach.tree import converters


class FieldKwargs[T](TypedDict, total=False):
    default: T
    validator: attrs._ValidatorArgType[T] | None
    repr: attrs._ReprArgType
    hash: bool | None
    init: bool
    metadata: Mapping[Any, Any] | None
    converter: (
        attrs._ConverterType
        | list[attrs._ConverterType]
        | tuple[attrs._ConverterType, ...]
        | None
    )
    factory: Callable[[], T] | None
    kw_only: bool | None
    eq: attrs._EqOrderType | None
    order: attrs._EqOrderType | None
    on_setattr: attrs._OnSetAttrArgType | None
    alias: str | None
    type: type | None


def array(**kwargs: Unpack[FieldKwargs[ArrayLike | None]]) -> Array:
    kwargs.setdefault("converter", converters.asarray)
    return field(**kwargs)  # pyright: ignore[reportReturnType]


@wraps(attrs.field)
def container(**kwargs) -> Any:
    if "converter" in kwargs and "factory" not in kwargs:
        kwargs["factory"] = kwargs["converter"]  # pyright: ignore[reportGeneralTypeIssues]
    elif "converter" not in kwargs and "factory" in kwargs:
        kwargs["converter"] = kwargs["factory"]  # pyright: ignore[reportGeneralTypeIssues]
    elif "converter" not in kwargs and "factory" not in kwargs:
        kwargs["converter"] = _dict_if_none
        kwargs["factory"] = dict
    return field(**kwargs)


@wraps(attrs.field)
def field(**kwargs) -> Any:
    if "default_factory" in kwargs:
        kwargs.setdefault("factory", kwargs.pop("default_factory"))
    if kwargs.pop("static", False):
        kwargs["metadata"] = toolz.assoc(kwargs.get("metadata") or {}, "static", True)  # noqa: FBT003
    return attrs.field(**kwargs)  # pyright: ignore[reportCallIssue]


@wraps(attrs.field)
def static(**kwargs) -> Any:
    kwargs.setdefault("static", True)
    return field(**kwargs)  # pyright: ignore[reportCallIssue]


@overload
def _dict_if_none(value: None) -> dict: ...
@overload
def _dict_if_none[T](value: T) -> T: ...
def _dict_if_none(value: Any) -> Any:
    if value is None:
        return {}
    return value



from collections.abc import Callable

import jax.numpy as jnp
from jaxtyping import Array, ArrayLike, Bool, Float

type Scalar = Float[Array, ""]
type Vector = Float[Array, " N"]


def absolute_residual(
    matvec: Callable[[Vector], Vector], x: Vector, b: Vector
) -> Scalar:
    r: Vector = matvec(x) - b
    r_norm: Scalar = jnp.linalg.norm(r)
    return r_norm


def relative_residual(
    matvec: Callable[[Vector], Vector], x: Vector, b: Vector
) -> Scalar:
    r_norm: Scalar = absolute_residual(matvec, x, b)
    b_norm: Scalar = jnp.linalg.norm(b)
    return r_norm / b_norm


def satisfies_tolerance(
    matvec: Callable[[Vector], Vector],
    x: Vector,
    b: Vector,
    *,
    atol: Float[ArrayLike, ""] = 1e-15,
    rtol: Float[ArrayLike, ""] = 1e-5,
) -> Bool[Array, ""]:
    residual: Scalar = absolute_residual(matvec, x, b)
    b_norm: Scalar = jnp.linalg.norm(b)
    return residual <= atol + rtol * b_norm



from collections.abc import Iterable
from typing import override

import jax.numpy as jnp
from jaxtyping import Array, Float

from liblaf.peach import tree
from liblaf.peach.constraints import Constraint
from liblaf.peach.linalg import utils
from liblaf.peach.linalg.abc import Callback, LinearSolution, LinearSolver, Result
from liblaf.peach.linalg.system import LinearSystem

type Scalar = Float[Array, ""]
type Vector = Float[Array, " free"]


def _default_solvers() -> list[LinearSolver]:
    from liblaf.peach import cuda

    if cuda.is_available():
        from liblaf.peach.linalg.cupy import CupyCG, CupyMinRes

        return [CupyCG(), CupyMinRes()]

    from liblaf.peach.linalg.scipy import ScipyCG, ScipyMinRes

    return [ScipyCG(), ScipyMinRes()]


@tree.define
class CompositeSolver(LinearSolver):
    from ._composite_types import CompositeState as State
    from ._composite_types import CompositeStats as Stats

    Solution = LinearSolution[State, Stats]

    solvers: list[LinearSolver] = tree.field(factory=_default_solvers)
    continue_atol: Float[Array, ""] = tree.array(
        default=0.0, converter=tree.converters.asarray, kw_only=True
    )
    continue_rtol: Float[Array, ""] = tree.array(
        default=1e-3, converter=tree.converters.asarray, kw_only=True
    )

    @override
    def _solve(
        self,
        system: LinearSystem,
        state: State,
        stats: Stats,
        *,
        constraints: Iterable[Constraint] = (),
        callback: Callback[State, Stats] | None = None,
    ) -> tuple[State, Stats, Result]:
        solution: LinearSolution = None  # pyright: ignore[reportAssignmentType]
        for solver in self.solvers:
            solution = solver.solve(
                system, state.params, callback=callback, constraints=constraints
            )
            state.state.append(solution.state)
            stats.stats.append(solution.stats)
            assert system.matvec is not None
            abs_residual: Scalar = utils.absolute_residual(
                system.matvec, solution.params_flat, system.b_flat
            )
            b_norm: Scalar = jnp.linalg.norm(system.b_flat)
            if abs_residual <= self.continue_atol + self.continue_rtol * b_norm:
                state.params_flat = solution.params_flat
            stats.relative_residual = abs_residual / b_norm
            if solution.success:
                break
        state.params_flat = solution.params_flat
        return state, stats, solution.result
